/*********************************************
 * vim:sw=8:ts=8:si:et
 * To use the above modeline in vim you must have "set modeline" in your .vimrc
 * Author: Guido Socher
 * Copyright: GPL V2
 * See http://www.gnu.org/licenses/gpl.html
 *
 * Ethernet remote device and sensor
 * HTTP interface 
 *
 * Chip type           : Atmega88 or Atmega168 or Atmega328 with ENC28J60
 * Note: there is a version number in the text. Search for tuxgraphics
 *********************************************/
#include <avr/io.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>
#include "ip_arp_udp_tcp.h"
#include "enc28j60.h"
#include "timeout.h"
#include "avr_compat.h"
#include "analog.h"
#include "net.h"

// please modify the following lines. mac and ip have to be unique
// in your local area network. You can not have the same numbers in
// two devices. The IP address may be changed at run-time but the
// MAC address is hardcoded:
static uint8_t mymac[6] = {0x54,0x55,0x58,0x10,0x00,0x29};
// how did I get the mac addr? Translate the first 3 numbers into ascii is: TUX
static uint8_t myip[4] = {10,0,0,29};
//static uint8_t myip[4] = {192,168,255,100};
// listen port for tcp/www (max range 1-254)
#define MYWWWPORT 80
// The buffer is the packet size we can handle and its upper limit is
// given by the amount of RAM that the atmega chip has.
#define BUFFER_SIZE 650
static uint8_t buf[BUFFER_SIZE+1];

static uint8_t modifyip=0;  // 2 means set it back to default, 1 means change
// global string buffer
#define STR_BUFFER_SIZE 24
static char strbuf[STR_BUFFER_SIZE+1];
//
//
// search for a string of the form key=value in
// a string that looks like q?xyz=abc&uvw=defgh HTTP/1.1\r\n
// The returned value is stored in the global var strbuf
uint8_t find_key_val(char *str,char *key)
{
        uint8_t found=0;
        uint8_t i=0;
        char *kp;
        kp=key;
        while(*str &&  *str!=' ' && found==0){
                if (*str == *kp){
                        kp++;
                        if (*kp == '\0'){
                                str++;
                                kp=key;
                                if (*str == '='){
                                        found=1;
                                }
                        }
                }else{
                        kp=key;
                }
                str++;
        }
        if (found==1){
                // copy the value to a buffer and terminate it with '\0'
                while(*str &&  *str!=' ' && *str!='&' && i<STR_BUFFER_SIZE){
                        strbuf[i]=*str;
                        i++;
                        str++;
                }
                strbuf[i]='\0';
        }
        // return the length of the value
        return(i);
}

// convert a single hex digit character to its integer value
unsigned char h2int(char c)
{
        if (c >= '0' && c <='9'){
                return((unsigned char)c - '0');
        }
        if (c >= 'a' && c <='f'){
                return((unsigned char)c - 'a' + 10);
        }
        if (c >= 'A' && c <='F'){
                return((unsigned char)c - 'A' + 10);
        }
        return(0);
}

// decode a url string e.g "hello%20joe" or "hello+joe" becomes "hello joe"
void urldecode(char *urlbuf)
{
        char c;
        char *dst;
        dst=urlbuf;
        while ((c = *urlbuf)) {
                if (c == '+') c = ' ';
                if (c == '%') {
                        urlbuf++;
                        c = *urlbuf;
                        urlbuf++;
                        c = (h2int(c) << 4) | h2int(*urlbuf);
                }
                *dst = c;
                dst++;
                urlbuf++;
        }
        *dst = '\0';
}

// parse a string an extract the IP to bytestr
uint8_t parse_ip(uint8_t *bytestr,char *str)
{
        char *sptr;
        uint8_t i=0;
        sptr=NULL;
        while(i<4){
                bytestr[i]=0;
                i++;
        }
        i=0;
        while(*str && i<4){
                // if a number then start
                if (sptr==NULL && isdigit(*str)){
                        sptr=str;
                }
                if (*str == '.'){
                        *str ='\0';
                        bytestr[i]=(atoi(sptr)&0xff);
                        i++;
                        sptr=NULL;
                }
                str++;
        }
        *str ='\0';
        if (i==3){
                bytestr[i]=(atoi(sptr)&0xff);
                return(0);
        }
        return(1);
}

// take a byte string and display it  (base is 10 for ip and 16 for mac addr)
void mk_net_str(char *resultstr,uint8_t *bytestr,uint8_t len,char separator,uint8_t base)
{
        uint8_t i=0;
        uint8_t j=0;
        while(i<len){
                itoa((int)bytestr[i],&resultstr[j],base);
                // search end of str:
                while(resultstr[j]){j++;}
                resultstr[j]=separator;
                j++;
                i++;
        }
        j--;
        resultstr[j]='\0';
}

void store_in_eeprom(void)
{
        eeprom_write_byte((uint8_t *)0x0,19); // magic number
        eeprom_write_block((uint8_t *)myip,(void *)1,sizeof(myip));
}

// return 2 in case of error:
int8_t store_new_ip(char *str)
{
        if (find_key_val(str,"nip")){
                urldecode(strbuf);
                if (parse_ip(myip,strbuf)==0){
                        store_in_eeprom();
                        return(1);
                }
        }
        return(2);
}

// takes a string of the form command/Number and analyse it (e.g "pd7/1 HTTP/1.1")
// return values: 0 invalid url data 
//                1 ok
int8_t analyse_get_url(char *str)
{
        uint8_t on=0;
        uint8_t port=0;
        // it must be pd followed by a number
        if (modifyip ==1 && *str == 'i'){
                return(store_new_ip(str));
        }
        if (*str != 'p'){
                return(0);
        }
        str++;
        if (*str != 'd'){
                return(0);
        }
        str++;
        if (*str < 0x3a && *str > 0x2f){
                // is a ASCII number, return it
                port=*str-0x30;
        }else{
                return(0);
        }
        str++;
        // now a slash
        if (*str != '/'){
                return(0);
        }
        str++;
        // now zero or 1
        if (*str == '1'){
                on=1;
        }
        str++;
        // now space
        if (*str != ' '){
                return(0);
        }
        if (port == 7){
                if (on){
                        PORTD|= (1<<PORTD7);// transistor on
                }else{
                        PORTD &= ~(1<<PORTD7);// transistor off
                }
        }
        if (port == 6){
                if (on){
                        PORTD|= (1<<PORTD6);// transistor on
                }else{
                        PORTD &= ~(1<<PORTD6);// transistor off
                }
        }
        if (port == 5){
                if (on){
                        PORTD|= (1<<PORTD5);// transistor on
                }else{
                        PORTD &= ~(1<<PORTD5);// transistor off
                }
        }
        if (port == 4){
                if (on){
                        PORTD|= (1<<PORTD4);// transistor on
                }else{
                        PORTD &= ~(1<<PORTD4);// transistor off
                }
        }
        return(1);
}


// modify own IP:
uint16_t print_webpage_config(uint8_t *buf)
{
        uint16_t plen;
        plen=fill_tcp_data_p(buf,0,PSTR("HTTP/1.0 200 OK\r\nContent-Type: text/html\r\nPragma: no-cache\r\n\r\n"));
        plen=fill_tcp_data_p(buf,plen,PSTR("<h2>Change web server IP</h2><pre>\n"));
        plen=fill_tcp_data_p(buf,plen,PSTR("<form action=/i method=get>"));
        plen=fill_tcp_data_p(buf,plen,PSTR("new IP: <input type=text size=12 name=nip value="));
        mk_net_str(strbuf,myip,4,'.',10);
        plen=fill_tcp_data(buf,plen,strbuf);
        plen=fill_tcp_data_p(buf,plen,PSTR("><br>\n"));
        plen=fill_tcp_data_p(buf,plen,PSTR("<input type=submit value=\"change\"></form><hr>"));
        return(plen);
}

uint16_t print_webpage_confirm(uint8_t *buf)
{
        uint16_t plen;
        plen=fill_tcp_data_p(buf,0,PSTR("HTTP/1.0 200 OK\r\nContent-Type: text/html\r\nPragma: no-cache\r\n\r\n"));
        plen=fill_tcp_data_p(buf,plen,PSTR("<h2>OK IP updated.</h2> Take out th jumper and power cycle.\n"));
        return(plen);
}

// prepare the webpage by writing the data to the tcp send buffer
uint16_t print_webpage(uint8_t *buf)
{
        uint16_t plen;
        char numstr[6];
        plen=fill_tcp_data_p(buf,0,PSTR("HTTP/1.0 200 OK\r\nContent-Type: text/html\r\nPragma: no-cache\r\n\r\n"));
        plen=fill_tcp_data_p(buf,plen,PSTR("<h2>Line status:</h2>"));
        plen=fill_tcp_data_p(buf,plen,PSTR("<pre>"));
        plen=fill_tcp_data_p(buf,plen,PSTR("\nPD4 output: "));
        if (PORTD & (1<<PORTD4)){
                plen=fill_tcp_data_p(buf,plen,PSTR("on  [<a href=/pd4/0>change</a>]"));
        }else{
                plen=fill_tcp_data_p(buf,plen,PSTR("off [<a href=/pd4/1>change</a>]"));
        }
        plen=fill_tcp_data_p(buf,plen,PSTR("\nPD5 output: "));
        if (PORTD & (1<<PORTD5)){
                plen=fill_tcp_data_p(buf,plen,PSTR("on  [<a href=/pd5/0>change</a>]"));
        }else{
                plen=fill_tcp_data_p(buf,plen,PSTR("off [<a href=/pd5/1>change</a>]"));
        }
        plen=fill_tcp_data_p(buf,plen,PSTR("\nPD6 output: "));
        if (PORTD & (1<<PORTD6)){
                plen=fill_tcp_data_p(buf,plen,PSTR("on  [<a href=/pd6/0>change</a>]"));
        }else{
                plen=fill_tcp_data_p(buf,plen,PSTR("off [<a href=/pd6/1>change</a>]"));
        }
        plen=fill_tcp_data_p(buf,plen,PSTR("\nPD7 output: "));
        if (PORTD & (1<<PORTD7)){
                plen=fill_tcp_data_p(buf,plen,PSTR("on  [<a href=/pd7/0>change</a>]"));
        }else{
                plen=fill_tcp_data_p(buf,plen,PSTR("off [<a href=/pd7/1>change</a>]"));
        }
        plen=fill_tcp_data_p(buf,plen,PSTR("\ndigital input PD0: "));
        if (bit_is_clear(PIND,PIND0)){
                plen=fill_tcp_data_p(buf,plen,PSTR("0"));
        }else{
                plen=fill_tcp_data_p(buf,plen,PSTR("1"));
        }
        plen=fill_tcp_data_p(buf,plen,PSTR("\ndigital input PD1: "));
        if (bit_is_clear(PIND,PIND1)){
                plen=fill_tcp_data_p(buf,plen,PSTR("0"));
        }else{
                plen=fill_tcp_data_p(buf,plen,PSTR("1"));
        }
        plen=fill_tcp_data_p(buf,plen,PSTR("\nanalog input ADC0: "));
        itoa(convertanalog(0),numstr,10); // convert integer to string
        plen=fill_tcp_data(buf,plen,numstr);
        plen=fill_tcp_data_p(buf,plen,PSTR("\nanalog input ADC1: "));
        itoa(convertanalog(1),numstr,10); // convert integer to string
        plen=fill_tcp_data(buf,plen,numstr);

        plen=fill_tcp_data_p(buf,plen,PSTR("\n\n<a href=\"/\">[refresh page]</a>\n</pre><hr>\n"));
        plen=fill_tcp_data_p(buf,plen,PSTR("<br><small>version 2.17, tuxgraphics.org<br>\n"));
        return(plen);
}


int main(void){

        
        uint16_t plen;
        uint16_t dat_p;
        int8_t cmd;
        
        // set the clock speed to "no pre-scaler" (8MHz with internal osc or 
        // full external speed)
        // set the clock prescaler. First write CLKPCE to enable setting of clock the
        // next four instructions.
        CLKPR=(1<<CLKPCE); // change enable
        CLKPR=0; // "no pre-scaler"
        _delay_loop_1(50); // 12ms

        /* enable PD2/INT0, as input */
        DDRD&= ~(1<<DDD2);

        // test PD3=reset to default 
        cbi(DDRD,PIND3);
        sbi(PORTD,PIND3); // internal pullup resistor on
        // test PD2=change IP
        cbi(DDRD,PIND2);
        sbi(PORTD,PIND2); // internal pullup resistor on

        _delay_loop_1(50); // 12ms
        if (bit_is_clear(PIND,PIND2)){
                modifyip=1;
        }
        if (bit_is_clear(PIND,PIND3)){
                // make eeprom data invalid
                eeprom_write_byte((uint8_t *)0x0,1); // delete magic number
        }
        _delay_loop_1(50); // 12ms
        if (eeprom_read_byte((uint8_t *)0x0) == 19){
                // ok magic number matches accept values
                eeprom_read_block((uint8_t *)myip,(void *)1,sizeof(myip));
        }
        /*initialize enc28j60*/
        enc28j60Init(mymac);
        enc28j60clkout(2); // change clkout from 6.25MHz to 12.5MHz
        _delay_loop_1(50); // 12ms
        //
        adc_ref_pin_init();
        
        // LED
        /* enable PB1, LED as output */
        //DDRB|= (1<<DDB1);
        /* set output to Vcc, LED off */
        //PORTB|= (1<<PORTB1);

        // the transistor on PD7
        DDRD|= (1<<DDD7);
        PORTD &= ~(1<<PORTD7);// transistor off
        // the transistor on PD6
        DDRD|= (1<<DDD6);
        PORTD &= ~(1<<PORTD6);// transistor off
        // the transistor on PD5
        DDRD|= (1<<DDD5);
        PORTD &= ~(1<<PORTD5);// transistor off
        // the transistor on PD4
        DDRD|= (1<<DDD4);
        PORTD &= ~(1<<PORTD4);// transistor off
        // digital input:
        DDRD&= ~(1<<DDD0);
        // pullup on
        PORTD|= (1<<PIND0);
        // digital input:
        DDRD&= ~(1<<DDD1);
        // pullup on
        PORTD|= (1<<PIND1);
        
        /* Magjack leds configuration, see enc28j60 datasheet, page 11 */
        // LEDB=yellow LEDA=green
        //
        // 0x476 is PHLCON LEDA=links status, LEDB=receive/transmit
        // enc28j60PhyWrite(PHLCON,0b0000 0100 0111 01 10);
        enc28j60PhyWrite(PHLCON,0x476);
        _delay_loop_1(50); // 12ms
        

        //init the ethernet/ip layer:
        init_ip_arp_udp_tcp(mymac,myip,MYWWWPORT);

        while(1){
                // get the next new packet:
                plen = enc28j60PacketReceive(BUFFER_SIZE, buf);

                /*plen will ne unequal to zero if there is a valid 
                 * packet (without crc error) */
                if(plen==0){
                        continue;
                }
                        
                // arp is broadcast if unknown but a host may also
                // verify the mac address by sending it to 
                // a unicast address.
                if(eth_type_is_arp_and_my_ip(buf,plen)){
                        make_arp_answer_from_request(buf);
                        continue;
                }

                // check if ip packets are for us:
                if(eth_type_is_ip_and_my_ip(buf,plen)==0){
                        continue;
                }
                
                if(buf[IP_PROTO_P]==IP_PROTO_ICMP_V && buf[ICMP_TYPE_P]==ICMP_TYPE_ECHOREQUEST_V){
                        // a ping packet, let's send pong
                        make_echo_reply_from_request(buf,plen);
                        continue;
                }
                // tcp port www start, compare only the lower byte
                if (buf[IP_PROTO_P]==IP_PROTO_TCP_V&&buf[TCP_DST_PORT_H_P]==0&&buf[TCP_DST_PORT_L_P]==MYWWWPORT){
                        if (buf[TCP_FLAGS_P] & TCP_FLAGS_SYN_V){
                                make_tcp_synack_from_syn(buf);
                                // make_tcp_synack_from_syn does already send the syn,ack
                                continue;
                        }
                        if (buf[TCP_FLAGS_P] & TCP_FLAGS_ACK_V){
                                init_len_info(buf); // init some data structures
                                // we can possibly have no data, just ack:
                                dat_p=get_tcp_data_pointer();
                                if (dat_p==0){
                                        if (buf[TCP_FLAGS_P] & TCP_FLAGS_FIN_V){
                                                // finack, answer with ack
                                                make_tcp_ack_from_any(buf);
                                        }
                                        // just an ack with no data, wait for next packet
                                        continue;
                                }
                                if (strncmp("GET ",(char *)&(buf[dat_p]),4)!=0){
                                        // head, post and other methods:
                                        //
                                        // for possible status codes see:
                                        // http://www.w3.org/Protocols/rfc2616/rfc2616-sec10.html
                                        plen=fill_tcp_data_p(buf,0,PSTR("HTTP/1.0 200 OK\r\nContent-Type: text/html\r\n\r\n<h1>200 OK</h1>"));
                                        goto SENDTCP;
                                }
                                if (strncmp("/ ",(char *)&(buf[dat_p+4]),2)==0){
                                        if (modifyip){
                                                plen=print_webpage_config(buf);
                                        }else{
                                                plen=print_webpage(buf);
                                        }
                                        goto SENDTCP;
                                }
                                // analyse the url and do possible port changes:
                                cmd=analyse_get_url((char *)&(buf[dat_p+5]));
                                // for possible status codes see:
                                // http://www.w3.org/Protocols/rfc2616/rfc2616-sec10.html
                                if (cmd==0){
                                        plen=fill_tcp_data_p(buf,0,PSTR("HTTP/1.0 404 Not Found\r\nContent-Type: text/html\r\n\r\n<h1>404 Not Found</h1>"));
                                        goto SENDTCP;
                                }
                                if (cmd==2){
                                        plen=fill_tcp_data_p(buf,0,PSTR("HTTP/1.0 406 Not Acceptable\r\nContent-Type: text/html\r\n\r\n<h1>406 IP addr. format wrong</h1>"));
                                        goto SENDTCP;
                                }
                                // just display the status:
                                if (modifyip){
                                        plen=print_webpage_confirm(buf);
                                }else{
                                        plen=print_webpage(buf);
                                }
                                //
SENDTCP:
                                make_tcp_ack_from_any(buf); // send ack for http get
                                make_tcp_ack_with_data(buf,plen); // send data
                                continue;
                        }

                }
                // tcp port www end
                //
        }
        return (0);
}
